import { __classPrivateFieldSet, __classPrivateFieldGet, __awaiter } from 'tslib';
import { Deferred, isPromiseLike, assert, convertTime, wrapMethod } from '@s-libs/js-core';
import { remove, nth, isEqual, functions, get, isObject, bindKey, clone, forOwn, isFunction, flatMap, keys } from '@s-libs/micro-dash';
import { HttpClientTestingModule, HttpTestingController } from '@angular/common/http/testing';
import { ApplicationRef, Component, InjectFlags } from '@angular/core';
import { tick, flush, TestBed, flushMicrotasks, discardPeriodicTasks, fakeAsync } from '@angular/core/testing';
import { HarnessEnvironment } from '@angular/cdk/testing';
import { UnitTestElement } from '@angular/cdk/testing/testbed';
import { NoopAnimationsModule } from '@angular/platform-browser/animations';
import { By } from '@angular/platform-browser';
import { TestScheduler } from 'rxjs/testing';
import { CompileMetadataResolver } from '@angular/compiler';
import { BrowserDynamicTestingModule, platformBrowserDynamicTesting } from '@angular/platform-browser-dynamic/testing';

/**
 * A mock method call that was made and is ready to be answered. This interface allows access to the underlying <code>jasmine.CallInfo</code>, and allows resolving or rejecting the asynchronous call's result.
 */
class TestCall {
    constructor(deferred, ctx) {
        this.deferred = deferred;
        this.ctx = ctx;
    }
    /**
     * Resolve the call with the given value.
     */
    flush(value) {
        var _a;
        this.deferred.resolve(value);
        (_a = this.ctx) === null || _a === void 0 ? void 0 : _a.tick();
    }
    /**
     * Reject the call with the given reason.
     */
    error(reason) {
        var _a;
        this.deferred.reject(reason);
        (_a = this.ctx) === null || _a === void 0 ? void 0 : _a.tick();
    }
}

var _spy, _testCalls;
/**
 * Controller to be used in tests, that allows for mocking and flushing any asynchronous function. For example, to mock the browser's paste functionality:
 *
 * ```ts
 *  it('can paste', () => {
 *   const clipboard = navigator.clipboard;
 *   const ctx = new AngularContextNext();
 *
 *   // mock the browser API for pasting
 *   const controller = new AsyncMethodController(clipboard, 'readText', {
 *     ctx,
 *   });
 *   ctx.run(() => {
 *     // BEGIN production code that copies to the clipboard
 *     let pastedText: string;
 *     clipboard.readText().then((text) => {
 *       pastedText = text;
 *     });
 *     // END production code that copies to the clipboard
 *
 *     // mock the behavior when the user denies access to the clipboard
 *     controller.expectOne([]).flush('mock clipboard contents');
 *
 *     // BEGIN expect the correct results after a successful copy
 *     expect(pastedText!).toBe('mock clipboard contents');
 *     // END expect the correct results after a successful copy
 *   });
 * });
 * ```
 */
class AsyncMethodController {
    /**
     * Optionally provide `ctx` to automatically trigger promise handlers and changed detection after calling `flush()` or `error()`. This is the normal production behavior of asynchronous browser APIs. However, if the function you are stubbing is not patched by zone.js, change detection would not run automatically, in which case you many not want to pass this parameter. See the list of functions that zone.js patches [here](https://github.com/angular/angular/blob/master/packages/zone.js/STANDARD-APIS.md).
     */
    constructor(obj, methodName, { ctx = undefined } = {}) {
        _spy.set(this, void 0);
        _testCalls.set(this, []);
        // Note: it wasn't immediately clear how avoid `any` in this constructor, and this will be invisible to users. So I gave up. (For now?)
        __classPrivateFieldSet(this, _spy, spyOn(obj, methodName));
        __classPrivateFieldGet(this, _spy).and.callFake((() => {
            const deferred = new Deferred();
            __classPrivateFieldGet(this, _testCalls).push(new TestCall(deferred, ctx));
            return deferred.promise;
        }));
    }
    /**
     * Expect that a single call has been made which matches the given parameters or predicate, and return its mock. If no such call has been made, or more than one such call has been made, fail with an error message including `description`, if provided.
     */
    expectOne(match, description) {
        const matches = this.match(match);
        if (matches.length !== 1) {
            throw new Error(this.buildErrorMessage({
                matchType: 'one matching',
                matches,
                stringifiedUserInput: this.stringifyUserInput(match, description),
            }));
        }
        return matches[0];
    }
    /**
     * Expect that no calls have been made which match the given parameters or predicate. If a matching call has been made, fail with an error message including `description`, if provided.
     */
    expectNone(match, description) {
        const matches = this.match(match);
        if (matches.length > 0) {
            throw new Error(this.buildErrorMessage({
                matchType: 'zero matching',
                matches,
                stringifiedUserInput: this.stringifyUserInput(match, description),
            }));
        }
    }
    /**
     * Search for calls that match the given parameters or predicate, without any expectations.
     */
    match(match) {
        this.ensureCallInfoIsSet();
        const filterFn = Array.isArray(match)
            ? this.makeArgumentMatcher(match)
            : match;
        return remove(__classPrivateFieldGet(this, _testCalls), (testCall) => filterFn(testCall.callInfo));
    }
    /**
     * Verify that no unmatched calls are outstanding. If any calls are outstanding, fail with an error message indicating which calls were not handled.
     */
    verify() {
        if (__classPrivateFieldGet(this, _testCalls).length) {
            this.ensureCallInfoIsSet();
            let message = this.buildErrorMessage({
                matchType: 'no open',
                matches: __classPrivateFieldGet(this, _testCalls),
            }) + ':';
            for (const testCall of __classPrivateFieldGet(this, _testCalls)) {
                message += `\n  ${stringifyArgs(testCall.callInfo.args)}`;
            }
            throw new Error(message);
        }
    }
    ensureCallInfoIsSet() {
        for (let i = 1; i <= __classPrivateFieldGet(this, _testCalls).length; ++i) {
            const testCall = nth(__classPrivateFieldGet(this, _testCalls), -i);
            if (testCall.callInfo) {
                return;
            }
            testCall.callInfo = nth(__classPrivateFieldGet(this, _spy).calls.all(), -i);
        }
    }
    makeArgumentMatcher(args) {
        return (callInfo) => isEqual(callInfo.args, args);
    }
    buildErrorMessage({ stringifiedUserInput, matchType, matches, }) {
        let message = `Expected ${matchType} call(s)`;
        if (stringifiedUserInput) {
            message += ` for criterion "${stringifiedUserInput}"`;
        }
        message += `, found ${matches.length}`;
        return message;
    }
    stringifyUserInput(match, description) {
        if (!description) {
            if (Array.isArray(match)) {
                description = 'Match by arguments: ' + stringifyArgs(match);
            }
            else {
                description = 'Match by function: ' + match.name;
            }
        }
        return description;
    }
}
_spy = new WeakMap(), _testCalls = new WeakMap();
/** @hidden */
function stringifyArgs(args) {
    return JSON.stringify(args);
}

// adapted from https://github.com/ngneat/spectator/blob/e13c9554778bdb179dfc7235aedb4b3b90302850/projects/spectator/src/lib/mock.ts
/**
 * Creates a new object with jasmine spies for each method in `type`.
 *
 * ```ts
 * class Greeter {
 *   greet(name: string) {
 *     return `Hello, ${name}!`;
 *   }
 * }
 *
 * const spyObject = createSpyObject(Greeter);
 * spyObject.greet.and.returnValue("Hello, stub!");
 * expect(spyObject.greet("Eric")).toBe("Hello, stub!");
 * expectSingleCallAndReset(spyObject.greet, "Eric");
 * ```
 */
function createSpyObject(type) {
    const mock = {};
    for (let proto = type.prototype; proto; proto = Object.getPrototypeOf(proto)) {
        for (const key of functions(proto)) {
            mock[key] = jasmine.createSpy(key);
        }
    }
    return mock;
}

/**
 * Expects the spy to have been called the exact number of times as the argument lists provided, with those arguments, then resets the spy.
 *
 * ```ts
 * const spy = jasmine.createSpy();
 *
 * spy(1)
 * spy(2, 3)
 * spy()
 * expectCallsAndReset(spy, [1], [2, 3], []);
 * ```
 */
function expectCallsAndReset(spy, ...allArgs) {
    expect(spy).toHaveBeenCalledTimes(allArgs.length);
    expect(spy.calls.allArgs()).toEqual(allArgs);
    spy.calls.reset();
}

/**
 * Expects exactly one call to have been made to a jasmine spy, for it to have received the given arguments, then resets the spy.
 *
 * ```ts
 * const spy = jasmine.createSpy();
 *
 * spy(1, 2);
 * expectSingleCallAndReset(spy, 1, 2); // pass
 * expectSingleCallAndReset(spy, 1, 2); // fail
 *
 * spy(3);
 * spy(4);
 * expectSingleCallAndReset(spy, 3); // fail
 * ```
 */
function expectSingleCallAndReset(spy, ...params) {
    expect(spy).toHaveBeenCalledTimes(1);
    expect(spy).toHaveBeenCalledWith(...params);
    spy.calls.reset();
}

/** @hidden */
const proxyTarget = Symbol('proxy target'); // trick from https://stackoverflow.com/a/53431924/1836506
/** @hidden */
function synchronize(obj) {
    return new Proxy(obj, {
        get(target, p, receiver) {
            if (p === proxyTarget) {
                return obj;
            }
            else {
                return synchronizeAny(Reflect.get(target, p, receiver));
            }
        },
        apply(target, thisArg, argArray) {
            thisArg = get(thisArg, [proxyTarget], thisArg);
            return synchronizeAny(Reflect.apply(target, thisArg, argArray));
        },
    });
}
/** @hidden */
function synchronizeAny(value) {
    if (isPromiseLike(value)) {
        return synchronizePromise(value);
    }
    else if (isObject(value)) {
        return synchronize(value);
    }
    else {
        return value;
    }
}
/** @hidden */
function synchronizePromise(promise) {
    let awaited;
    promise.then((value) => (awaited = value));
    tick();
    // TODO: a nice error if the promise didn't resolve
    return synchronizeAny(awaited);
}

/** @hidden */
class FakeAsyncHarnessEnvironment extends HarnessEnvironment {
    constructor(rawRootElement, ctx) {
        super(rawRootElement);
        this.ctx = ctx;
    }
    static documentRootLoader(ctx) {
        return synchronize(new FakeAsyncHarnessEnvironment(document.body, ctx));
    }
    waitForTasksOutsideAngular() {
        return __awaiter(this, void 0, void 0, function* () {
            flush();
        });
    }
    forceStabilize() {
        return __awaiter(this, void 0, void 0, function* () {
            this.ctx.tick();
        });
    }
    createEnvironment(element) {
        return new FakeAsyncHarnessEnvironment(element, this.ctx);
    }
    createTestElement(element) {
        return new UnitTestElement(element, bindKey(this, 'forceStabilize'));
    }
    getAllRawElements(selector) {
        return __awaiter(this, void 0, void 0, function* () {
            return Array.from(this.rawRootElement.querySelectorAll(selector));
        });
    }
    getDocumentRoot() {
        return document.body;
    }
}

/** @hidden */
function extendMetadata(metadata, toAdd) {
    const result = clone(metadata);
    forOwn(toAdd, (val, key) => {
        result[key] = Array.isArray(result[key]) ? result[key].concat(val) : val;
    });
    return result;
}
/**
 * @deprecated Use {@link AngularContextNext} instead. This old version will be removed in a future version, and `AngularContextNext` will be renamed to `AngularContext`.
 *
 * Some notes to migrate to the new version:
 * - For any tests that were passing inputs to `.run()`, instead call `.assignInputs()` before `.run()`.
 */
class AngularContext {
    /**
     * @param moduleMetadata passed along to [TestBed.configureTestingModule()]{@linkcode https://angular.io/api/core/testing/TestBed#configureTestingModule}. Automatically includes {@link HttpClientTestingModule} for you.
     */
    constructor(moduleMetadata = {}) {
        /**
         * Set this before calling `run()` to mock the time at which the test starts.
         */
        this.startTime = new Date();
        this.loader = FakeAsyncHarnessEnvironment.documentRootLoader(this);
        TestBed.configureTestingModule(extendMetadata(moduleMetadata, { imports: [HttpClientTestingModule] }));
    }
    run(optionsOrTest, test) {
        let options = {};
        if (isFunction(optionsOrTest)) {
            test = optionsOrTest;
        }
        else {
            options = optionsOrTest;
        }
        this.runWithMockedTime(() => {
            assert(test);
            this.init(options);
            try {
                test();
                this.verifyPostTestConditions();
            }
            finally {
                this.cleanUp();
            }
        });
    }
    /**
     * Gets a service or other injectable from the root injector. This implementation is a simple pass-through to [TestBed.inject()]{@linkcode https://angular.io/api/core/testing/TestBed#inject}, but subclasses may provide their own implementation. It is recommended to use this in your tests instead of using `TestBed` directly.
     */
    inject(token) {
        return TestBed.inject(token);
    }
    /**
     * Gets a component harness, wrapped for use in a fakeAsync test so that you do not need to `await` its results. Throws an error if no match can be located.
     */
    getHarness(query) {
        return this.loader.getHarness(query);
    }
    /**
     * Gets a component harness, wrapped for use in a fakeAsync test so that you do not need to `await` its results. Returns `null` if the harness cannot be located.
     */
    getHarnessForOptional(query) {
        return this.loader.locatorForOptional(query)();
    }
    /**
     * Gets all component harnesses that match the query, wrapped for use in a fakeAsync test so that you do not need to `await` its results.
     */
    getAllHarnesses(query) {
        return this.loader.getAllHarnesses(query);
    }
    /**
     * Advance time and trigger change detection. It is common to call this with no arguments to trigger change detection without advancing time.
     *
     * @param unit The unit of time `amount` represents. Accepts anything described in `@s-libs/s-core`'s [TimeUnit]{@linkcode https://simontonsoftware.github.io/s-js-utils/typedoc/enums/timeunit.html} enum.
     */
    tick(amount = 0, unit = 'ms') {
        // To simulate real life, trigger change detection before processing macro tasks. To further simulate real life, wait until the micro task queue is empty.
        flushMicrotasks();
        this.runChangeDetection();
        tick(convertTime(amount, unit, 'ms'));
        this.runChangeDetection();
    }
    /**
     * This is a hook for subclasses to override. It is called during `run()`, before the `test()` callback. This implementation does nothing, but if you override this it is still recommended to call `super.init(options)` in case this implementation does something in the future.
     */
    init(_options) { }
    /** @hidden */
    runChangeDetection() {
        this.inject(ApplicationRef).tick();
    }
    /**
     * Runs post-test verifications. This base implementation runs [HttpTestingController#verify]{@linkcode https://angular.io/api/common/http/testing/HttpTestingController#verify}. Unlike {@link #cleanUp}, it is OK for this method to throw an error to indicate a violation.
     */
    verifyPostTestConditions() {
        this.inject(HttpTestingController).verify();
    }
    /**
     * Performs any cleanup needed at the end of each test. This base implementation calls [discardPeriodicTasks]{@linkcode https://angular.io/api/core/testing/discardPeriodicTasks} to avoid an error from the `fakeAsync` zone.
     */
    cleanUp() {
        discardPeriodicTasks();
    }
    runWithMockedTime(test) {
        // https://github.com/angular/angular/issues/31677#issuecomment-573139551
        const now = performance.now;
        spyOn(performance, 'now').and.callFake(() => Date.now());
        jasmine.clock().install();
        fakeAsync(() => {
            jasmine.clock().mockDate(this.startTime);
            test();
        })();
        jasmine.clock().uninstall();
        performance.now = now;
    }
}

/** @hidden */
function extendMetadata$1(metadata, toAdd) {
    const result = clone(metadata);
    forOwn(toAdd, (val, key) => {
        result[key] = Array.isArray(result[key]) ? result[key].concat(val) : val;
    });
    return result;
}
/**
 * Provides the foundation for an opinionated testing pattern.
 * - All test are run in the `fakeAsync` zone. This gives you full control over
 *   the timing of everything by default.
 * - Variables that are initialized for each test exist in a context that is
 *   thrown away, so they cannot leak between tests.
 * - Clearly separates initialization code from the test itself.
 * - Gives control over the simulated date & time with a single line of code.
 * - Automatically includes {@link HttpClientTestingModule} to stub network requests without additional setup.
 * - Always verifies no unexpected http requests were made during a test.
 * - Always discards periodic tasks at the end of each test to automatically
 *   avoid an error from the `fakeAsync` zone.
 *
 * Why does the class name end with "Next"? This replaces the old `AngularContext`, but it's a breaking change so this gives people some time to transition over. Eventually the old one will be removed and this will be renamed to `AngularContext`.
 *
 * This example tests a simple service that uses HttpClient, and is tested by using `AngularContextNext` directly. More often `AngularContextNext` will be used a super class. See {@link ComponentContextNext} for more common use cases.
 *
 * ```ts
 * // This is the class we will test.
 * @Injectable({ providedIn: 'root' })
 * class MemoriesService {
 *   constructor(private httpClient: HttpClient) {}
 *
 *   getLastYearToday(): Observable<any> {
 *     const datetime = new Date();
 *     datetime.setFullYear(datetime.getFullYear() - 1);
 *     const date = datetime.toISOString().split('T')[0];
 *     return this.httpClient.get(`http://example.com/post-from/${date}`);
 *   }
 * }
 *
 * describe('MemoriesService', () => {
 *   // Tests should have exactly 1 variable outside an "it": `ctx`.
 *   let ctx: AngularContextNext;
 *   beforeEach(() => {
 *     ctx = new AngularContextNext();
 *   });
 *
 *   it('requests a post from 1 year ago', () => {
 *     // Before calling `run`, set up any context variables this test needs.
 *     ctx.startTime = new Date('2004-02-16T10:15:00.000Z');
 *
 *     // Pass the test itself as a callback to `run()`.
 *     ctx.run(() => {
 *       const httpBackend = ctx.inject(HttpTestingController);
 *       const myService = ctx.inject(MemoriesService);
 *
 *       myService.getLastYearToday().subscribe();
 *
 *       httpBackend.expectOne('http://example.com/post-from/2003-02-16');
 *     });
 *   });
 * });
 * ```
 */
class AngularContextNext {
    /**
     * @param moduleMetadata passed along to [TestBed.configureTestingModule()]{@linkcode https://angular.io/api/core/testing/TestBed#configureTestingModule}. Automatically includes {@link HttpClientTestingModule} for you.
     */
    constructor(moduleMetadata = {}) {
        /**
         * Set this before calling `run()` to mock the time at which the test starts.
         */
        this.startTime = new Date();
        TestBed.configureTestingModule(extendMetadata$1(moduleMetadata, { imports: [HttpClientTestingModule] }));
    }
    /**
     * Runs `test` in a `fakeAsync` zone. It can use async/await, but be sure anything you `await` is already due to execute (e.g. if a timeout is due in 3 seconds, call `.tick(3000)` before `await`ing its result).
     *
     * Also runs the following in this order, all within the same zone:
     *
     * 1. `this.init()`
     * 2. `test()`
     * 3. `this.verifyPostTestConditions()`
     * 4. `this.cleanUp()`
     */
    run(test) {
        this.runWithMockedTime(() => {
            this.init();
            try {
                test();
                this.tick();
                this.verifyPostTestConditions();
            }
            finally {
                this.cleanUp();
            }
        });
    }
    /**
     * Gets a service or other injectable from the root injector. This implementation is a simple pass-through to [TestBed.inject()]{@linkcode https://angular.io/api/core/testing/TestBed#inject}, but subclasses may provide their own implementation. It is recommended to use this in your tests instead of using `TestBed` directly.
     */
    inject(token) {
        return TestBed.inject(token);
    }
    /**
     * Advance time and trigger change detection. It is common to call this with no arguments to trigger change detection without advancing time.
     *
     * @param unit The unit of time `amount` represents. Accepts anything described in `@s-libs/s-core`'s [TimeUnit]{@linkcode https://simontonsoftware.github.io/s-js-utils/typedoc/enums/timeunit.html} enum.
     */
    tick(amount = 0, unit = 'ms') {
        // To simulate real life, trigger change detection before processing macro tasks. To further simulate real life, wait until the micro task queue is empty.
        flushMicrotasks();
        this.runChangeDetection();
        tick(convertTime(amount, unit, 'ms'));
        this.runChangeDetection();
    }
    /**
     * This is a hook for subclasses to override. It is called during `run()`, before the `test()` callback. This implementation does nothing, but if you override this it is still recommended to call `super.init()` in case this implementation does something in the future.
     */
    init() { }
    /** @hidden */
    runChangeDetection() {
        this.inject(ApplicationRef).tick();
    }
    /**
     * Runs post-test verifications. This base implementation runs [HttpTestingController#verify]{@linkcode https://angular.io/api/common/http/testing/HttpTestingController#verify}. Unlike {@link #cleanUp}, it is OK for this method to throw an error to indicate a violation.
     */
    verifyPostTestConditions() {
        this.inject(HttpTestingController).verify();
    }
    /**
     * Performs any cleanup needed at the end of each test. This base implementation calls [discardPeriodicTasks]{@linkcode https://angular.io/api/core/testing/discardPeriodicTasks} to avoid an error from the `fakeAsync` zone.
     */
    cleanUp() {
        discardPeriodicTasks();
    }
    runWithMockedTime(test) {
        // https://github.com/angular/angular/issues/31677#issuecomment-573139551
        const now = performance.now;
        spyOn(performance, 'now').and.callFake(() => Date.now());
        jasmine.clock().install();
        fakeAsync(() => {
            jasmine.clock().mockDate(this.startTime);
            test();
        })();
        jasmine.clock().uninstall();
        performance.now = now;
    }
}

/** @hidden */
let initialStyles;
/**
 * When called at the right times, this speeds up your test suite by removing `<script>` tags that Angular component tests leave in the DOM. By default Angular dynamically adds `<script>` tags into the `<head>` of the document corresponding to component styles. In Jasmine tests they are never cleaned up. This can cause hundreds or even thousands of style tags to accumulate, making component tests slow down as your test suite runs. In Chrome debugger's performance tab, this time will show up under the activity "Recalculate Style".
 *
 * The first time this function is called it memorizes the `<script>` tags in the DOM, then each time it is called after it removes any that appeared since. A simple way to use this effectively is to call it in a `beforeEach()`.
 *
 * ```ts
 * beforeEach(() => {
 *   trimLeftoverStyles();
 * });
 * ```
 */
function trimLeftoverStyles() {
    const styles = new Set(Array.from(document.querySelectorAll('style')));
    if (initialStyles) {
        for (const style of styles) {
            if (!initialStyles.has(style)) {
                style.remove();
            }
        }
    }
    else {
        initialStyles = styles;
    }
}
/** @hidden */
function resetTrimLeftoverStyles() {
    initialStyles = undefined;
}

/**
 * @deprecated Use {@link ComponentContextNext} instead. This old version will be removed in a future version, and `ComponentContextNext` will be renamed to `ComponentContext`.
 *
 * Some notes to migrate to the new version:
 * - Change "input" to "inputs" in calls to `.run({ input: {...}}, ...)`
 * - Any variables you set in "inputs" must now be actual angular `@Input()`s
 * - `ctx.fixture` is now for a synthetic wrapper component, not your component. Change `ctx.fixture.componentInstance` to `ctx.getComponentInstance()`
 * - If your component has an input with a default value that you want to preserve, pass it in the constructor's `unboundInputs` argument.
 */
class ComponentContext extends AngularContext {
    /**
     * @param moduleMetadata passed along to [TestBed.configureTestingModule()]{@linkcode https://angular.io/api/core/testing/TestBed#configureTestingModule}. Automatically includes {@link NoopAnimationsModule}, in addition to those provided by {@link AngularContext}.
     */
    constructor(moduleMetadata = {}) {
        super(extendMetadata(moduleMetadata, { imports: [NoopAnimationsModule] }));
    }
    /**
     * Constructs and initializes your component. Called during `.run()` before it executes the rest of your test. Runs in the same `fakeAsync` zone as the rest of your test.
     */
    init(options) {
        trimLeftoverStyles();
        super.init(options);
        this.fixture = TestBed.createComponent(this.componentType);
        Object.assign(this.fixture.componentInstance, options.input);
        this.fixture.detectChanges();
        this.tick();
    }
    /** @hidden */
    runChangeDetection() {
        this.fixture.detectChanges();
    }
    /**
     * Performs any cleanup needed at the end of each test. This implementation destroys {@link fixture} and calls the super implementation.
     */
    cleanUp() {
        this.fixture.destroy();
        super.cleanUp();
    }
}

/** @hidden */
class FakeAsyncHarnessEnvironmentNext extends HarnessEnvironment {
    constructor(rawRootElement, ctx) {
        super(rawRootElement);
        this.ctx = ctx;
    }
    static documentRootLoader(ctx) {
        return new FakeAsyncHarnessEnvironmentNext(document.body, ctx);
    }
    waitForTasksOutsideAngular() {
        return __awaiter(this, void 0, void 0, function* () {
            flush();
        });
    }
    forceStabilize() {
        return __awaiter(this, void 0, void 0, function* () {
            this.ctx.tick();
        });
    }
    createEnvironment(element) {
        return new FakeAsyncHarnessEnvironmentNext(element, this.ctx);
    }
    createTestElement(element) {
        return new UnitTestElement(element, bindKey(this, 'forceStabilize'));
    }
    getAllRawElements(selector) {
        return __awaiter(this, void 0, void 0, function* () {
            return Array.from(this.rawRootElement.querySelectorAll(selector));
        });
    }
    getDocumentRoot() {
        return document.body;
    }
}

/** @hidden */
function createDynamicWrapper(componentType, unboundInputs) {
    const selector = getSelector(componentType);
    const inputMetas = getInputMetas(componentType).filter(({ property }) => !unboundInputs.includes(property));
    class DynamicWrapperComponent {
        constructor() {
            this.inputs = {};
            this.styles = {};
        }
    }
    DynamicWrapperComponent.decorators = [
        { type: Component, args: [{ template: buildTemplate(selector, inputMetas) },] }
    ];
    const type = DynamicWrapperComponent;
    const inputProperties = inputMetas.map((meta) => meta.property);
    return { type, inputProperties };
}
/** @hidden */
function getSelector(componentType) {
    var _a;
    const annotations = Reflect.getOwnPropertyDescriptor(componentType, '__annotations__');
    assert(annotations, 'That does not appear to be a component');
    let selector = (_a = annotations.value.find((decorator) => decorator.selector)) === null || _a === void 0 ? void 0 : _a.selector;
    if (!isValidSelector(selector)) {
        selector = 's-libs-component-under-test';
        TestBed.overrideComponent(componentType, { set: { selector } });
    }
    return selector;
}
/** @hidden */
function isValidSelector(selector) {
    if (!selector) {
        return false;
    }
    try {
        document.createElement(selector);
        return true;
    }
    catch (_a) {
        return false;
    }
}
/** @hidden */
function getInputMetas(componentType) {
    // I tried making this support inputs with special characters in their names, but it turns out that *Angular* can only support that when using AOT. So our *dynamic* wrapper cannot.
    return flatMap(componentType.propDecorators, (decorators, property) => {
        const inputDecorators = decorators.filter((decorator) => decorator.type.prototype.ngMetadataName === 'Input');
        return inputDecorators.map((decorator) => {
            var _a;
            const binding = ((_a = decorator.args) === null || _a === void 0 ? void 0 : _a[0]) || property;
            return { property, binding };
        });
    });
}
/** @hidden */
function buildTemplate(selector, inputMetas) {
    const bindingStrings = inputMetas.map(({ binding, property }) => `[${binding}]="inputs.${property}"`);
    return `
    <div class="s-libs-dynamic-wrapper" [ngStyle]="styles">
      <${selector} ${bindingStrings.join(' ')}></${selector}>
    </div>
  `;
}

/**
 * A superclass to set up testing contexts for components. This is a foundation for an opinionated testing pattern, including everything described in {@link AngularContextNext} plus:
 *
 * - Automatically creates your component at the beginning of `run()`.
 * - Wraps your component in a dynamically created parent component. (This sets up Angular to call `ngOnChanges()` in your tests the same way it does in production.)
 * - Lets you use {@link https://material.angular.io/cdk/test-harnesses/overview|component harnesses} (which are normally not usable within a `fakeAsync` test).
 * - Automatically disables animations.
 * - Automatically integrates {@link trimLeftoverStyles} to speed up your test suite.
 *
 * Why does the class name end with "Next"? This replaces the old `ComponentContext`, but it's a breaking change so this gives people some time to transition over. Eventually the old one will be removed and this will be renamed to `ComponentContext`.
 *
 * A very simple example:
 * ```ts
 * @Component({ template: 'Hello, {{name}}!' })
 * class GreeterComponent {
 *   @Input() name!: string;
 * }
 *
 * it('greets you by name', () => {
 *   const ctx = new ComponentContextNext(GreeterComponent);
 *   ctx.assignInputs({ name: 'World' });
 *   ctx.run(() => {
 *     expect(ctx.fixture.nativeElement.textContent).toBe('Hello, World!');
 *   });
 * });
 * ```
 *
 * A full example, with routing and a component harness. This is the full code for a tiny Angular app:
 * ```ts
 * /////////////////
 * // app-context.ts
 *
 * // To re-use your context setup, make a subclass of ComponentContext to
 * // import into any spec
 * class AppContext extends ComponentContextNext<AppComponent> {
 *   constructor() {
 *     super(AppComponent, {
 *       imports: [
 *         // This is your production `AppModule`. Make sure it exports `AppComponent`
 *         AppModule,
 *         // Import `routes` from your `app-routing.module.ts`
 *         RouterTestingModule.withRoutes(routes),
 *       ],
 *     });
 *   }
 * }
 *
 * ////////////////////////
 * // app.component.spec.ts
 *
 * describe('AppComponent', () => {
 *   let ctx: AppContext;
 *   beforeEach(() => {
 *     ctx = new AppContext();
 *   });
 *
 *   it('can navigate to the first page', () => {
 *     ctx.run(() => {
 *       ctx.getHarness(AppComponentHarness).navigateToFirstPage();
 *       expect(ctx.fixture.nativeElement.textContent).toContain(
 *         'First works!',
 *       );
 *     });
 *   });
 * });
 *
 * ///////////////////////////
 * // app.component.harness.ts
 *
 * // A simple component harness to demonstrate its integration with component contexts. Note that everything here is asynchronous, but is treated as synchronous in tests. This is a trick in AngularContexts so harnesses can be used in the fakeAsync zone.
 * class AppComponentHarness extends ComponentHarness {
 *   static hostSelector = 'app-root';
 *
 *   private getFirstPageLink = this.locatorFor('a');
 *
 *   async navigateToFirstPage(): Promise<void> {
 *     const link = await this.getFirstPageLink();
 *     await link.click();
 *   }
 * }
 *
 * /////////////////////
 * // first.component.ts
 *
 * // A minimal component for demonstration purposes
 * @Component({ template: '<p>First works!</p>' })
 * class FirstComponent {}
 *
 * ///////////////////
 * // app.component.ts
 *
 * // A minimal app component with routing for demonstration purposes
 * @Component({
 *   selector: 'app-root',
 *   template: `
 *     <a routerLink="/first-page">First Page</a>
 *     <router-outlet></router-outlet>
 *   `,
 * })
 * class AppComponent {}
 *
 * ////////////////////////
 * // app-routing.module.ts
 *
 * // exported for use in tests
 * const routes: Routes = [{ path: 'first-page', component: FirstComponent }];
 *
 * @NgModule({
 *   imports: [RouterModule.forRoot(routes)],
 *   exports: [RouterModule],
 * })
 * class AppRoutingModule {}
 *
 * ////////////////
 * // app.module.ts
 *
 * // A minimal app module. Notice the added export.
 * @NgModule({
 *   declarations: [AppComponent, FirstComponent],
 *   imports: [BrowserModule, BrowserAnimationsModule, AppRoutingModule],
 *   bootstrap: [AppComponent],
 *   exports: [AppComponent], // exported for use in tests
 * })
 * class AppModule {}
 * ```
 */
class ComponentContextNext extends AngularContextNext {
    /**
     * @param componentType `run()` will create a component of this type before running the rest of your test.
     * @param moduleMetadata passed along to [TestBed.configureTestingModule()]{@linkcode https://angular.io/api/core/testing/TestBed#configureTestingModule}. Automatically includes {@link NoopAnimationsModule}, in addition to those provided by {@link AngularContextNext}.
     * @param unboundInputs By default a synthetic parent component will be created that binds to all your component's inputs. Pass input names here that should NOT be bound. This is useful e.g. to test the default value of an input.
     */
    constructor(componentType, moduleMetadata = {}, unboundInputs = []) {
        const wrapper = createDynamicWrapper(componentType, unboundInputs);
        super(extendMetadata(moduleMetadata, {
            imports: [NoopAnimationsModule],
            declarations: [wrapper.type, componentType],
        }));
        this.componentType = componentType;
        this.wrapperType = wrapper.type;
        this.inputProperties = new Set(wrapper.inputProperties);
        this.inputs = {};
        this.wrapperStyles = {};
    }
    /**
     * Assign css styles to the div wrapping your component. Can be called before or during `run()`. Accepts an object with the same structure as the {@link https://angular.io/api/common/NgStyle|ngStyle directive}.
     *
     * ```ts
     * ctx.assignWrapperStyles({
     *   width: '400px',
     *   height: '600px',
     *   margin: '20px auto',
     *   border: '1px solid',
     * });
     * ```
     */
    assignWrapperStyles(styles) {
        Object.assign(this.wrapperStyles, styles);
        if (this.isInitialized()) {
            this.flushStylesToWrapper();
            this.tick();
        }
    }
    /**
     * Assign inputs passed into your component. Can be called before `run()` to set the initial inputs, or within `run()` to update them and trigger all the appropriate change detection and lifecycle hooks.
     */
    assignInputs(inputs) {
        for (const key of keys(inputs)) {
            if (!this.inputProperties.has(key)) {
                throw new Error(`Cannot bind to "${key}" (it is not an input, or you passed it in \`unboundProperties\`)`);
            }
        }
        Object.assign(this.inputs, inputs);
        if (this.isInitialized()) {
            this.flushInputsToWrapper();
            this.tick();
        }
    }
    /**
     * Use within `run()` to get your instantiated component that is on the page.
     */
    getComponentInstance() {
        return this.fixture.debugElement.query(By.directive(this.componentType))
            .componentInstance;
    }
    /**
     * Gets a component harness, wrapped for use in a fakeAsync test so that you do not need to `await` its results. Throws an error if no match can be located.
     */
    getHarness(query) {
        return this.loader.getHarness(query);
    }
    /**
     * Gets all component harnesses that match the query, wrapped for use in a fakeAsync test so that you do not need to `await` its results.
     */
    getAllHarnesses(query) {
        return this.loader.getAllHarnesses(query);
    }
    /**
     * Constructs and initializes your component. Called during `run()` before it executes the rest of your test. Runs in the same `fakeAsync` zone as the rest of your test.
     */
    init() {
        trimLeftoverStyles();
        super.init();
        this.fixture = TestBed.createComponent(this.wrapperType);
        this.loader = FakeAsyncHarnessEnvironmentNext.documentRootLoader(this);
        this.flushStylesToWrapper();
        this.flushInputsToWrapper();
        this.fixture.detectChanges();
        this.tick();
    }
    /** @hidden */
    runChangeDetection() {
        this.fixture.detectChanges();
    }
    /**
     * Performs any cleanup needed at the end of each test. This implementation destroys {@link fixture} and calls the super implementation.
     */
    cleanUp() {
        this.fixture.destroy();
        super.cleanUp();
    }
    isInitialized() {
        return !!this.fixture;
    }
    flushInputsToWrapper() {
        Object.assign(this.getWrapperComponentInstance().inputs, this.inputs);
    }
    flushStylesToWrapper() {
        Object.assign(this.getWrapperComponentInstance().styles, this.wrapperStyles);
    }
    getWrapperComponentInstance() {
        return this.fixture.componentInstance;
    }
}

/**
 * Designed for temporary use when a test using `fakeAsync` is giving you "Error: X timer(s) still in the queue", and you don't know what they are. Call this at the top of the test to see everything that adds a timer printed to the JS console. In Chrome you can click the function in the console output to be taken directly to its definition in the Sources tab.
 *
 * Be sure to call this _inside_ the test, not before, to get the most useful output.
 *
 * @returns a function to stop the logging
 *
 * ```ts
 * it('definitely does not set any timeouts', fakeAsync(() => {
 *   const stopLogging = logTimers();
 *
 *   doTheRestOfMyTest();
 *
 *   stopLogging();
 * }));
 * ```
 */
function logTimers() {
    const restoreTimeouts = wrapMethod(window, 'setTimeout', {
        before(...args) {
            console.log('setTimeout(', ...args, ')');
        },
    });
    const restoreIntervals = wrapMethod(window, 'setInterval', {
        before(...args) {
            console.log('setInterval(', ...args, ')');
        },
    });
    return () => {
        restoreTimeouts();
        restoreIntervals();
    };
}

/**
 * A helper to take some boilerplate out of RxJS's new marble testing API for jasmine.
 *
 * ```ts
 * it("doubles numbers", marbleTest(({ cold, expectObservable }) => {
 *   const source = cold("-1-2-3-|");
 *   const expected = "   -2-4-6-|");
 *   expectObservable(source.pipe(double())).toBe(expected);
 * }));
 * ```
 */
function marbleTest(callback) {
    const testScheduler = new TestScheduler((actual, expected) => {
        expect(actual).toEqual(expected);
    });
    return () => testScheduler.run((helpers) => callback(Object.assign(Object.assign({}, helpers), { testScheduler })));
}

/**
 * Use this to speed up your angular test suite. Normally when you call `TestBed.compileComponents()`, it recompiles all the configured components even if it has compiled them before in a previous test. AoT compilation would be a great fix, but is not well supported/documented for tests yet. This function fills the gap by precompiling the components as if by AoT and setting up `TestBed` to use them.
 *
 * ```ts
 * // let's assume `AppModule` declares or imports a `HelloWorldComponent`
 * precompileForTests([AppModule]);
 *
 * // Everything below here is the same as normal. Just add the line above.
 *
 * describe("AppComponent", () => {
 *   it("says hello", async () => {
 *     TestBed.configureTestingModule({ declarations: [HelloWorldComponent] });
 *     await TestBed.compileComponents(); // <- this line is faster
 *     const fixture = TestBed.createComponent(HelloWorldComponent);
 *     expect(fixture.nativeElement.textContent).toContain("Hello, world!");
 *   });
 * });
 * ```
 *
 * Note that this uses `TestBed.initTestEnvironment` to set up the precompiled components, so you will not be able to use it yourself at the same time.
 *
 * @deprecated This is no longer needed with Ivy.
 */
function precompileForTests(modules, skipModules = []) {
    beforeAll(() => __awaiter(this, void 0, void 0, function* () {
        // technique modeled from https://github.com/angular/angular/blob/11325bad4ab786a07e52ff380c00622fda11c0b7/packages/core/test/linker/jit_summaries_integration_spec.ts#L83
        const metadataResolver = optionallyInject(CompileMetadataResolver);
        if (!metadataResolver) {
            console.warn(`precompileForTests() is no longer needed with Ivy. Skipping.`);
            return;
        }
        TestBed.configureTestingModule({ imports: modules });
        yield TestBed.compileComponents();
        TestBed.resetTestEnvironment();
        TestBed.initTestEnvironment(BrowserDynamicTestingModule, platformBrowserDynamicTesting(), () => extractAotSummaries(metadataResolver, modules, new Set(skipModules)));
    }));
}
// TODO: this could be a public function
/** @hidden */
function optionallyInject(token) {
    return TestBed.inject(token, null, InjectFlags.Optional);
}
/** @hidden */
function extractAotSummaries(metadataResolver, modules, skipModules) {
    return modules.map((module) => () => extractAotSummaries0(metadataResolver, module, skipModules));
}
/** @hidden */
function extractAotSummaries0(metadataResolver, module, skipModules) {
    if (skipModules.has(module)) {
        return [];
    }
    skipModules.add(module);
    const moduleMetadata = metadataResolver.getNgModuleMetadata(module);
    return [
        metadataResolver.getNgModuleSummary(module),
        () => extractDirectives(moduleMetadata, metadataResolver),
        () => extractPipes(moduleMetadata, metadataResolver),
        () => extractAotSummaries(metadataResolver, extractImports(moduleMetadata), skipModules),
    ];
}
/** @hidden */
function extractDirectives(moduleMetadata, metadataResolver) {
    return moduleMetadata.declaredDirectives.map((directive) => metadataResolver.getDirectiveSummary(directive.reference));
}
/** @hidden */
function extractPipes(moduleMetadata, metadataResolver) {
    return moduleMetadata.declaredPipes.map((pipe) => metadataResolver.getPipeSummary(pipe.reference));
}
/** @hidden */
function extractImports(moduleMetadata) {
    return moduleMetadata.importedModules.map((importedModule) => importedModule.type.reference);
}

/*
 * Public API Surface of ng-dev
 */

/**
 * Generated bundle index. Do not edit.
 */

export { AngularContext, AngularContextNext, AsyncMethodController, ComponentContext, ComponentContextNext, TestCall, createSpyObject, expectCallsAndReset, expectSingleCallAndReset, logTimers, marbleTest, precompileForTests, trimLeftoverStyles };
//# sourceMappingURL=s-libs-ng-dev.js.map
