/// <reference types="jasmine" />
import { TestCall } from './test-call';
/** @hidden */
declare type AsyncFunc = (...args: any[]) => Promise<any>;
/** @hidden */
declare type Match<WrappingObject, FunctionName extends AsyncMethodKeys<WrappingObject>> = Parameters<WrappingObject[FunctionName]> | ((callInfo: jasmine.CallInfo<WrappingObject[FunctionName]>) => boolean);
/** @hidden */
declare type AsyncMethodKeys<T> = {
    [k in keyof T]: T[k] extends AsyncFunc ? k : never;
}[keyof T];
/**
 * Controller to be used in tests, that allows for mocking and flushing any asynchronous function. For example, to mock the browser's paste functionality:
 *
 * ```ts
 *  it('can paste', () => {
 *   const clipboard = navigator.clipboard;
 *   const ctx = new AngularContextNext();
 *
 *   // mock the browser API for pasting
 *   const controller = new AsyncMethodController(clipboard, 'readText', {
 *     ctx,
 *   });
 *   ctx.run(() => {
 *     // BEGIN production code that copies to the clipboard
 *     let pastedText: string;
 *     clipboard.readText().then((text) => {
 *       pastedText = text;
 *     });
 *     // END production code that copies to the clipboard
 *
 *     // mock the behavior when the user denies access to the clipboard
 *     controller.expectOne([]).flush('mock clipboard contents');
 *
 *     // BEGIN expect the correct results after a successful copy
 *     expect(pastedText!).toBe('mock clipboard contents');
 *     // END expect the correct results after a successful copy
 *   });
 * });
 * ```
 */
export declare class AsyncMethodController<WrappingObject, FunctionName extends AsyncMethodKeys<WrappingObject>> {
    #private;
    /**
     * Optionally provide `ctx` to automatically trigger promise handlers and changed detection after calling `flush()` or `error()`. This is the normal production behavior of asynchronous browser APIs. However, if the function you are stubbing is not patched by zone.js, change detection would not run automatically, in which case you many not want to pass this parameter. See the list of functions that zone.js patches [here](https://github.com/angular/angular/blob/master/packages/zone.js/STANDARD-APIS.md).
     */
    constructor(obj: WrappingObject, methodName: FunctionName, { ctx }?: {
        ctx?: {
            tick(): void;
        } | undefined;
    });
    /**
     * Expect that a single call has been made which matches the given parameters or predicate, and return its mock. If no such call has been made, or more than one such call has been made, fail with an error message including `description`, if provided.
     */
    expectOne(match: Match<WrappingObject, FunctionName>, description?: string): TestCall<WrappingObject[FunctionName]>;
    /**
     * Expect that no calls have been made which match the given parameters or predicate. If a matching call has been made, fail with an error message including `description`, if provided.
     */
    expectNone(match: Match<WrappingObject, FunctionName>, description?: string): void;
    /**
     * Search for calls that match the given parameters or predicate, without any expectations.
     */
    match(match: Match<WrappingObject, FunctionName>): TestCall<WrappingObject[FunctionName]>[];
    /**
     * Verify that no unmatched calls are outstanding. If any calls are outstanding, fail with an error message indicating which calls were not handled.
     */
    verify(): void;
    private ensureCallInfoIsSet;
    private makeArgumentMatcher;
    private buildErrorMessage;
    private stringifyUserInput;
}
export {};
